@@ -8,7 +8,7 @@ package Graph::Easy::Group::Cell;
 use Graph::Easy::Node;
 
 @ISA = qw/Graph::Easy::Node/;
-$VERSION = 0.11;
+$VERSION = 0.12;
 
 use strict;
 
@@ -168,6 +168,11 @@ sub _set_label
   $self->{name} = $self->{group}->label();
   }
 
+sub shape
+  {
+  'rect';
+  }
+
 sub attribute
   {
   my $self = shift;
