@@ -8,13 +8,15 @@ package Graph::Easy::Layout::Grid;
 
 use vars qw/$VERSION/;
 
-$VERSION = '0.06';
+$VERSION = 0.06;
 
 #############################################################################
 #############################################################################
 
 package Graph::Easy;
 
+use strict;
+
 sub _balance_sizes
   {
   # Given a list of column/row sizes and a minimum size that their sum must
